#include "shared.h"
#include <cmath>
#include "analysisFunction.h"

/*
  Class that works with counts
*/

int calcSum(suint *counts,int len){
  int tmp=0;
  for(int i=0;i<len;i++)
    tmp += counts[i];
  return tmp;
}

double sum(const double *ary,int len);//<-from soapmaster.cpp


#define NUMBASES 4
//minor and major, is not matrix, but simply the adr of a might be null pointer
void phat1(suint **counts,int nFiles,double eps,int nSites,double *phats,int *major,int *minor,int *keepSites) {

  //algorithm goes on by a site on site basis
  
  double pis[nFiles];
  double wis[nFiles];
  
  //this is tricky, we shouldnt loop through the entire matrix, but just the number of lines that were actually read into the matrix
  //for (aMap::iterator it=asso.begin();it!=asso.end();it++){
  for(int s=0;s<nSites;s++){
    if(keepSites==0)
      continue;
  //    fprintf(oFile,"%s\t%d\t",it->first.chromo,it->first.position);
    //part one
    //first lets get the sum of each nucleotide
    int bases[NUMBASES] = {0,0,0,0};
    for(int i=0;i<nFiles;i++)
      for(size_t j=0;j<NUMBASES;j++){
	bases[j] += counts[s][i*NUMBASES+j];
	//	fprintf(oFile,"%d\t",bases[j]);
      }


    //now get the major/minor
    
    int majorA = 0;
    for(int i=1;i<4;i++)
      if (bases[i]>bases[majorA])
        majorA = i;
    //maj is now the major allele (most frequent)
    
    int temp=0;
    int minorA= majorA;
    for(int i=0;i<4;i++){
      if(majorA==i) //we should check the against the major allele      
        continue;
      else if (bases[i]>temp){
        minorA = i;
        temp=bases[i];
      }
    }
    major[s] = majorA;
    minor[s] = minorA;

    if(major[s]==minor[s]){
      phats[s] = 0.0;
      continue;
    }
    //if the site is variable
    

    for(int i=0;i<nFiles;i++){
      int ni = counts[s][i*4+minorA];
      int nt= counts[s][i*4+minorA] + counts[s][i*4+majorA]; 
      if(nt==0){//if we dont have any reads for individual 'i'
        pis[i] = 0;
        wis[i] = 0;

        continue;
      }
      pis[i] = (ni-eps*nt)/(nt*(1-2*eps));
      wis[i] = 2.0*nt/(nt+1.0);
    }
    
    double tmp=0;
    for(int i=0;i<nFiles;i++)
      tmp+= (pis[i]*wis[i]);
    
    double phat = std::max(0.0,tmp/sum(wis,nFiles));
    phats[s] = phat;
  }

}





void printCounts(loci *locs,suint **counts,int nSites,size_t nInd,FILE *oFileCountsPos,FILE *oFileCountsBin,int dumpType,int *keepSites){
  //  fprintf(stderr,"oFileCountsBin=%p\t oFileCountsPos=%p",oFileCountsBin,oFileCountsPos);

  for(int s=0;s<nSites;s++){
    //always print position and overall depth
    if(keepSites[s]==0)
      continue;
    fprintf(oFileCountsPos,"%s\t%d\t%d\n",locs[s].chromo,locs[s].position+1,calcSum(counts[s],4*nInd));
    
    //if we need per sample info
    if(oFileCountsBin) {
      if(dumpType==3)//count A,C,G,T
	for(int i=0;i<4*nInd;i++)
	  fprintf(oFileCountsBin,"%u\t",counts[s][i]);
      else{//print A+C+G+T
	for(int n=0;n<nInd;n++)
	  fprintf(oFileCountsBin,"%u\t",counts[s][n*4]+counts[s][n*4+1]+counts[s][n*4+2]+counts[s][n*4+3]);


      }
      fprintf(oFileCountsBin,"\n");	
    }
  }
  fflush(oFileCountsPos);
  if(dumpType>1)
    fflush(oFileCountsBin);
}



void printPhat(loci *locs,suint **counts,int *major,int *minor,double *phat,int nSites,size_t nInd,FILE *fp,int *keepSites) {
  //  fprintf(stderr,"oFileCountsBin=%p\t oFileCountsPos=%p",oFileCountsBin,oFileCountsPos);

  for(int s=0;s<nSites;s++){
    //always print position and overall depth
    if(keepSites[s]==0)
      continue;
    int bases[]={0,0,0,0};
    int nInfo =0;
    for(int i=0;i<nInd;i++){
      int dep=0;
      for(int j=0;j<4;j++){
	dep += counts[s][i*4+j];
	bases[j] += counts[s][i*4+j];
      }
      if(dep)
	nInfo++;
    }
    fprintf(fp,"%s\t%d\t%d\t%d\t",locs[s].chromo,locs[s].position+1,major[s],minor[s]);
    for(int j=0;j<4;j++)
      fprintf(fp,"%d\t",bases[j]);
    
    fprintf(fp,"%f\t%d\n",phat[s],nInfo);	
  }
  
  fflush(fp);

}




class countCls{
private:
  int dumpCounts;
  int doPhat;
  int doMajorMinor;
  int doCounts;
  int doQsDist;//0=nothing 1=overall 2=perbam
  //
  int qs_cutoff;
  float eps;
  FILE *oFileCountsBin;
  FILE *oFileCountsPos;
  FILE *oFileQs;
  FILE *oFilePhat;
  size_t *qsDist;
  
  //
  float cutoff;
  int minKeepIndC;
public:

  
  //none optional stuff
  
  countCls(const char *outfiles,argStruct arguments,int inputtype);
  ~countCls();
  void getOptions(argStruct arguments);
  void run(funkyPars  *pars);
  void printAndClean(funkyPars *pars);

};


void countCls::getOptions(argStruct arguments){
  //default
  cutoff = 0.005;// filter sites with phat 0.5% variability
  dumpCounts =0;
  doPhat = 0;
  eps = 0.001;
  doMajorMinor =  0;
  doCounts = 0;
  doQsDist = 0;
  minKeepIndC = -1;
  //
  qsDist = new size_t[256];
  memset(qsDist,0,256*sizeof(size_t));
  //from command line
  qs_cutoff =0;
  dumpCounts=angsd::getArg("-dumpCounts",dumpCounts,arguments);
  doCounts=angsd::getArg("-doCounts",doCounts,arguments);
  doPhat=angsd::getArg("-doPhat",doPhat,arguments);
  qs_cutoff=angsd::getArg("-qs",qs_cutoff,arguments);
  fprintf(stderr,"qs cutoff=%d\n",qs_cutoff);
  doQsDist=angsd::getArg("-doQsDist",doQsDist,arguments);
  eps=angsd::getArg("-eps",eps,arguments);
  cutoff=angsd::getArg("-cutoff",cutoff,arguments);
  doMajorMinor=angsd::getArg("-doMajorMinor",doMajorMinor,arguments);
  minKeepIndC = angsd::getArg("-minKeepIndC",minKeepIndC,arguments);

  if(dumpCounts&&doCounts==0){
    fprintf(stderr,"You must supply -doCounts if you want to dumpcounts\n");
    exit(0);
  }

  if(doPhat==1 && doMajorMinor!=2){
    fprintf(stderr,"You want to estimate phat, you need to select also -doMajorMinor 2\n");
    exit(0);
  }
  if(doQsDist){
    fprintf(stderr,"You should only use nThread 1 og nThreads 2, when summing QS (due to multiprogramming issues)\n");
    
  }
}

countCls::countCls(const char *outfiles,argStruct arguments,int inputtype){

  getOptions(arguments);

  //make output files
  const char* postfix2=".counts";
  const char* postfix1=".pos";
  const char *postfix3=".qs";
  const char *postfix4=".phat";

  oFileCountsPos = oFileCountsBin = oFileQs = oFilePhat = NULL;
  fprintf(stderr,"-dumpCounts=%d\tdoPhat=%d\teps=%f\tdoCounts=%d\tdoQsDist=%dcutoff=%f\n",dumpCounts,doPhat,eps,doCounts,doQsDist,cutoff);
  if(dumpCounts){
    oFileCountsPos=openFile(outfiles,postfix1);
    if(dumpCounts>1)
      oFileCountsBin = openFile(outfiles,postfix2);
  }
  if(doPhat)
    oFilePhat = openFile(outfiles,postfix4);
  if(doQsDist)
    oFileQs = openFile(outfiles,postfix3);
}


void printQs(FILE *fp,size_t *ary){
  for(int i=0;i<256;i++)
    fprintf(fp,"%d\t%lu\n",i,ary[i]);
  
}


countCls::~countCls(){
  //  fprintf(stderr,"[%s] calling destructor\n",__FUNCTION__);
  //  fflush(stderr);
  if(oFileCountsBin)
    fclose(oFileCountsBin);
  if(oFileCountsPos)
    fclose(oFileCountsPos);
  if(doQsDist){
    printQs(oFileQs,qsDist);
    fclose(oFileQs);
  }
  if(oFilePhat!=NULL)
    fclose(oFilePhat);
}


void countCls::printAndClean(funkyPars *pars){

  if(dumpCounts)
    printCounts(pars->sites,pars->counts,pars->numSites,pars->nInd,oFileCountsPos,oFileCountsBin,dumpCounts,pars->keepSites);

  if(oFilePhat)
    printPhat(pars->sites,pars->counts,pars->major,pars->minor,pars->phat,pars->numSites,pars->nInd,oFilePhat,pars->keepSites);
}


suint **countNucs(const chunkyT *chk,int qs_cutoff){
  //  fprintf(stderr,"[%s]\n",__FUNCTION__);
  
  suint **cnts = new suint*[chk->nSites];
  for(int s=0;s<chk->nSites;s++){//loop over sites
    //    fprintf(stderr,"nsamples=%d\n",chk->nSamples);
    cnts[s] = new suint[4*chk->nSamples];
    memset(cnts[s],0,4*chk->nSamples*sizeof(suint));
    for(int n=0;n<chk->nSamples;n++){//loop over samples
      for(int l=0;l<chk->nd[s][n].l;l++){
	//fprintf(stderr,"%c",chk->nd[s][n].seq[l]);
	if(chk->nd[s][n].seq[l]!='n' && chk->nd[s][n].seq[l]!='N' && chk->nd[s][n].qs[l]>=qs_cutoff)
	//if( chk->nd[s][n].qs[l]>=qs_cutoff)
	  cnts[s][4*n+refToInt[chk->nd[s][n].seq[l]]]++;
      }
    }
  }
  return cnts;
}


void countQs(const chunkyT *chk,size_t *ret){
  //  fprintf(stderr,"[%s]\n",__FUNCTION__);
  
  suint **cnts = new suint*[chk->nSites];
  for(int s=0;s<chk->nSites;s++){//loop over sites
    for(int n=0;n<chk->nSamples;n++){//loop over samples
      for(int l=0;l<chk->nd[s][n].l;l++){
	ret[chk->nd[s][n].qs[l]]++;
	//if( chk->nd[s][n].qs[l]>=qs_cutoff)
      }
    }
  }
}




void countCls::run(funkyPars *pars){
  if(doCounts||dumpCounts||doPhat){
    if(pars->chk==NULL||pars->counts!=NULL){
      fprintf(stderr,"chk is NULL or counts is not null, in either case something should be decided\n");
      exit(0);
    }

    if(doCounts||doPhat||dumpCounts||doMajorMinor==2)
      pars->counts = countNucs(pars->chk,qs_cutoff);
    else
      return;
    if(doPhat||doMajorMinor==2){
      if(pars->phat==NULL)
	pars->phat = new double [pars->numSites]; 
      if(pars->major==NULL)
	pars->major = new int [pars->numSites]; 
      if(pars->minor==NULL)
	pars->minor = new int [pars->numSites]; 
      phat1(pars->counts,pars->nInd,eps,pars->numSites,pars->phat,pars->major,pars->minor,pars->keepSites);
      //modify keepsites;
      for(int i=0;i<pars->numSites;i++){
	if(pars->phat[i]<cutoff)
	  pars->keepSites[i] = 0;
	
	//THIS IS REALLY STUPID but lets count number of samples wiht info
	int nDep =0;
	for(int s=0;s<pars->nInd;s++){
	  int dep=0;
	  for(int j=0;j<4;j++)
	    dep += pars->counts[i][s*4+j];
	  if(dep)
	    nDep++;
	}
	//nDep is now the number of sapmles wiht info
	if(minKeepIndC!=-1 && nDep<minKeepIndC)
	  pars->keepSites[i] = 0;

      }
    }
    
  }
  if(doQsDist)
    countQs(pars->chk,qsDist);
    
}
