/*
  copyright (gpl)   23 april 2012.
  Thorfinn Sand Korneliussen thorfinn@binf.ku.dk
  Anders albrechtsen albrecht@binf.ku.dk 
  0.12 more object oriented version of 0.03
  0.2 implemented bambi the native bamreader and GL estimator
*/

#include<iostream>

#include<cstring> //for cstring functions
#include<map> //associative array
#include<vector> //for the filenames
#include<cstdlib> //for exit()
#include<cstdio>
#include <sys/stat.h> //for checking for file existence.
#include <utility>
#include <zlib.h>
#include <signal.h>

#include "shared.h"
#include "parseArgs.h"
#include "multiReader.h"

#define VERSION 0.209
#define ARGS ".arg"

#include "angsd.h"
#include "bambi_interface.h"
#include "bammer.h"
extern std::vector <char *> dumpedFiles;//a vector of outputfilenames from shared.cpp


void printTime(){

  time_t t;
  time(&t);
  std::cerr<<ctime(&t);
}

extern "C"{
#include "angsd_bam_plcmd.h"
  loc *alloc_locs(int len);
  double *alloc_double(int len);
  double **alloc_double_pointer(int len);
  int *alloc_int(int len);
  int **alloc_int_pointer(int len);
}


loc *alloc_locs(int len){
  return new loc[len];
}

int *alloc_int(int len){
  int *tmp = new int[len];
  for(int i=0;i<len;i++)
    tmp[i] = 0;
  return tmp;
}

int **alloc_int_pointer(int len){
  int **tmp = new int*[len];
  return tmp;
}


double *alloc_double(int len){
  double *tmp = new double[len];
  for(int i=0;i<len;i++)
    tmp[i] = 0.0;
  return tmp;
}

double **alloc_double_pointer(int len){
  double **tmp = new double*[len];
  for(int i=0;i<len;i++)
    tmp[i] = NULL;
  return tmp;
}

progArgs *pg_glob = NULL;


/*
  this is the main function for pluggin in options need for the run
  This function will also plug in the reference and the ancestral allele
 */



void attachPars(funkyPars *fp,progArgs *pg){
  if(fp==NULL || pg==NULL){
    fprintf(stderr,"[%s] NULL pointer huge error\n",__FUNCTION__);
    exit(0);
  }
  fp->errors = pg->errors;

  //attach extra data
   if(pg->sim1!=NULL)
    fp->nInd =pg->nInd;
  else if(pg->beagle==NULL)
    fp->nInd= pg->filenames.size();

  fp->inputtype = pg->inputtype;
}


void printLoci(loci &l, FILE *fp){
  fprintf(fp,"%s\t%d\t",l.chromo,l.position);
}

void callBack(chunk *chk){
  funkyPars *fp = allocFunkyPars();
  fp->likes = chk->lik;
  fp->depth = chk->depth;
  fp->counts=chk->counts;

  fp->sites =(loci*) chk->locs;
  fp->numSites=chk->nSites;

  attachPars(fp,pg_glob);//remember to put after assining for loci*
  fp->nInd = chk->nInd;  //to set nInd after attachpars

  fp->chunkNumber = chk->chunkNumber;
  free(chk);
  selector(fp);//send data to shared/pml.cpp

}


void callBack_bambi(fcb *fff){
  //  fprintf(stderr,"[%s] fff=%p\n",__FUNCTION__, fff);
  //  fprintf(stderr,"[%s] fff->dn=%p\n",__FUNCTION__, fff->dn);
  if(fff==NULL)
    selector(NULL);//<-send NULL which acts as a killswitch
  else{
    //  fprintf(stderr,"in the callback of angsd.cpp\n");
    fflush(stderr);
    funkyPars *fp = allocFunkyPars();
    fp->for_callback=fff;
    attachPars(fp,pg_glob);//remember to put after assining for loci*
    
    //  fp->chunkNumber = chk->chunkNumber;
    //  free(chk);
    selector(fp);//send data to shared/pml.cpp
  }
}



int SIG_COND =1;//if we catch signal then quit program nicely
int VERBOSE =1;
extern pthread_mutex_t mUpPile_mutex;//mutex for not deleting mUppile data untill end of data
void handler(int s) {
  if(VERBOSE)
    fprintf(stderr,"Caught SIGNAL: %d will try to exit nicely (no more threads are created, we will wait for the current threads to finish)\n",s);
  VERBOSE=0;
  SIG_COND=0;
  pthread_mutex_unlock(&mUpPile_mutex);
}

//int bambi_main(int argc,char **argv) ;


int main(int argc, char** argv){
  clock_t t=clock();//how long time does the run take
  time_t t2=time(NULL);


  //we are threading so we want make a nice signal handler for ctrl+c
  struct sigaction sa;
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = handler;
  sigaction(SIGPIPE, &sa, 0);
  sigaction(SIGINT, &sa, 0);  



  //print arguments supplied
  for(int i=0;i<argc;i++)
    fprintf(stderr,"%s ",argv[i]);
  fprintf(stderr,"\n\t-> angsd version: %.3f\t build(%s %s)\n",VERSION,__DATE__,__TIME__); 
  printTime();  
  //initiate struct with argument information
  argStruct arguments;
  arguments.argc=argc;
  arguments.argv=argv;
  arguments.usedArgs= new int[argc];
  for(int i=0;i<argc;i++)
    arguments.usedArgs[i]=0;


  if(argc==1){//if haven't been supplied with arguments, load default,print, and exit
    progArgs *pg = allocProgArgs();
    printProgArgs(pg ,stderr);
    return 0;
  }

  progArgs *pg = parseArgs(argc,argv,arguments);
  fprintf(stderr,"nTrheads=%d\n",pg->nThreads);

  pg_glob = pg;

  fflush(stderr);
  fprintf(stderr,"\r\t-> ");

  

  init(pg->outfiles,pg->nThreads,pg->inputtype,arguments);

  parseArgStruct(arguments);//validate that all arguments has indeed been parsed in the instantiation 'init' above

  //print out the full arguments used for run.
  FILE *argFp = openFile(pg->outfiles,ARGS);
  for(int i=0;i<argc;i++)
    fprintf(argFp,"%s ",argv[i]);
  printProgArgs(pg,argFp);


  //instantiate our filereaders
  if(pg->mpileup==0 &&pg->uppile==0){
    multiReader mr(pg->inputtype,pg->target,pg->lStart,pg->lStop,pg->filenames,pg->faifile,pg->outfiles,pg->qs,pg->maxHits,pg->eps,pg->strand,pg->minDepth,pg->maxDepth,pg->cutOff,pg->downsample,pg->posfile,pg->nInd,arguments);
    /*
      In the above, everything from/including "outfiles"
      is only used for soap files
    */
    
    //    int chunkNumber =1;//we need a counter for our chunks
    while(SIG_COND) {
      funkyPars *tmp = mr.fetch(pg->nLines,pg->chunkSize); //YES MISTER FETCH

      for(int i=0;0&&i<250;i++)
	fprintf(stdout,"%f\t",tmp->likes[0][i]);
      //      return 0;
      if(tmp==NULL)
	break;  
      attachPars(tmp,pg);//plugin our options.
      //    fprintf(stdout,"mr.fetch nInd %lu \n",tmp->nInd);
      //tmp->chunkNumber = chunkNumber ++;//increment the chunknumber
      selector(tmp);//send data to shared/pml.cpp
      //fprintf(stderr,"\t-> tmp->nInd=%d\n",tmp->nInd);
      //      break;
    }
  }else if(pg->mpileup!=0){
    //    fprintf(stderr,"Will use samtools filereading\n");
    fprintf(stderr,"Deprecated\n");
    //    dirty_bam_main(argc-pg->mpileup,argv+pg->mpileup,callBack,pg->chunkSize,pg->qs);
  }else if(pg->uppile!=0){
    fprintf(stderr,"will use bambi filereading\n");
    bambi_main(argc-pg->uppile,argv+pg->uppile,pg->nThreads,arguments);
  }

  
  fprintf(stderr,"\n\r\t-> Done reading data waiting for calculations to finish\n");

  destroy();

  //print out nice messages


  fprintf(stderr, "\t[ALL done] cpu-time used =  %.2f sec\n", (float)(clock() - t) / CLOCKS_PER_SEC);
  fprintf(stderr, "\t[ALL done] walltime used =  %.2f sec\n", (float)(time(NULL) - t2));  
  fflush(stderr);
  fprintf(argFp, "\t[ALL done] cpu-time used =  %.2f sec\n", (float)(clock() - t) / CLOCKS_PER_SEC);
  fprintf(argFp, "\t[ALL done] walltime used =  %.2f sec\n", (float)(time(NULL) - t2));  
  fflush(argFp);


  
  killProgArgs(pg);
  
  //printout the filenames generated
  fprintf(stderr,"\t Below are the dumped filenames:\n");
  for(int i=0;i<(int)dumpedFiles.size();i++){
    fprintf(stderr,"\t\"%s\"\t",dumpedFiles[i]);
    fprintf(argFp,"\t\"%s\"\t",dumpedFiles[i]);
    free(dumpedFiles[i]);
  }
  fprintf(stderr,"\n");
  fprintf(argFp,"\n");
  printTime();
  fclose(argFp);
  delete []   arguments.usedArgs;
  return 0;
}
