#include "shared.h"
#include "multiReader.h"


multiReader::multiReader(int type_a,char *target_a,char *lStart,char *lStop,std::vector<char *> &filenames,char *faifile,char *outfiles,int qs,int maxHits,float eps,int strand,int minDepth,int maxDepth,float cutOff,int downsample,const char*posfile,int nInd,argStruct arguments) { // dragon. lots to clean up and move to read classes
  sm=NULL;
  pl=NULL;
  mytglfs=NULL;
  type = type_a;
  // pg=pg_in;

  target = NULL;  
  if(target_a!=NULL)
    target = buildTarget(target_a);

  switch(type){
  case 0:{
    fprintf(stderr,"\t[%s] assuming soapfile\n",__FUNCTION__);
    sm = new soapMaster (outfiles,target,filenames,qs,maxHits,eps,strand,minDepth,maxDepth,cutOff,downsample);
    break;
  }
  case 3:{
    fprintf(stderr,"\t[%s] assuming tglf\n",__FUNCTION__);
    mytglfs = new tglfs;
    mytglfs->init(filenames,posfile);
    break;
  }
  case 4:{
    fprintf(stderr,"\t[%s] assuming simulation files (single pop)\n",__FUNCTION__);
    mysims = new reader_sim;
    mysims->init(filenames[0],nInd);
    break;
  }
  case 5:{
    fprintf(stderr,"\t[%s] assuming beagle gprobs input \n",__FUNCTION__);
    bglObj = new beagle_reader;
    bglObj->init(filenames[0],arguments);
    break;
  }

  default:{
    fprintf(stderr,"\t[%s] assuming SAMtools pileup\n",__FUNCTION__);
    pl = new pileups (faifile,lStart,lStop,filenames,type);
    break;
  }
  }
  
}

multiReader::~multiReader(){
 
  switch(type){
  case 0:{
    delete sm;
    break;
  }
  case 3:{
    mytglfs->close();
    delete mytglfs;
    break;
  }
  case 4:{
    mysims->close();
    delete mysims;
    break;
  }    
  case 5:{
    delete bglObj;
      //    delete mysims;
    break;
  }    
  default:{
    delete pl;
    break;
  }
  }
  
  /*
    DRAGON should remove all internal structures in target
   */
  delete target;
}
funkyPars *multiReader::fetch(int nLines,int chunkSize){
    switch(type){
    case 0:{
      return sm->fetch(nLines,chunkSize);
      break;
    }
    case 3:{
      return mytglfs->fetch(nLines);
      break;
    }
    case 4:{
      funkyPars *fp = mysims->fetch(chunkSize); 
      //delete [] fp->anc;
      return fp;
      break;
    }
    case 5:{
      funkyPars *fp = bglObj->fetch(nLines,chunkSize); 
      //delete [] fp->anc;
      return fp;
      break;
    }
    default:{
      return pl->fetch(nLines,chunkSize);
      break;
    }
    }
    


}

oMap *multiReader::buildTarget(const char *filename){
  fprintf(stderr,"[%s] start indexing target region",__FUNCTION__);
  const char* delims = "\n \t:-;,.";
  oMap *asso = new oMap;


  char buffer[LENS];
  // std::ifstream pFile(filename,std::ios::in);
  FILE *pFile = fopen(filename,"r");
  
  char* chr = NULL;
  char* start;
  char* end;

  while(fgets(buffer,LENS,pFile)){
  //while(pFile.getline(buffer,LENS)){
    chr = strtok(buffer,delims);
    oMap::iterator it=asso->find(chr);
    iMap *tmpImap=NULL;
    if(it!=asso->end()){
      tmpImap = it->second;
    }else{
      tmpImap = new iMap;
      asso->insert(std::pair<char*,iMap*>(strdup(chr),tmpImap)) ;
    }
    
    start = strtok(NULL,delims); //start pos
    end = strtok(NULL,delims); //end pos
    if(atoi(end)<atoi(start)){
      fprintf(stderr,"\t-> Problem with target file, endpoint is before startpoint (%s,%s)\n",start,end);
      exit(0);
    }
    
    for(int i=atoi(start); i <atoi(end)+1;i++)
      tmpImap->insert(std::pair<int,int>(i,1));

  }
  //pFile.close();
  fclose(pFile);
  size_t sizes =0;
  for(oMap::const_iterator ot=asso->begin();ot!=asso->end();++ot){
    sizes += ot->second->size();
  }
  fprintf(stderr,"[%s] done indexing nSites=%lu\n",__FUNCTION__,sizes);
  fprintf(stderr,"[%s] =%p\n",__FUNCTION__,asso);
  return asso;
}
