/*
  assumes zero index positions

 */

#include "analysisFunction.h"

#include "shared.h"
#include "faidx.h"
//#include "getRef.h"



char *getRef1(const loci *locs,int lens,const char *fasta,int doCleanup){
  static char *chrLast =NULL;
  static char *ref = NULL;
  static faidx_t *fai = NULL;
  static int ref_len = -1;
  if(doCleanup){
    if(fai!=NULL)
      fai_destroy(fai);
    free(chrLast);
    free(ref);
    return NULL;
  }


  //  fprintf(stderr,"\t-> Reading fasta ref/anc [%s] fname=%s\n",__FUNCTION__,fasta);

  char *ret=new char[lens];

  
  
  if(chrLast==NULL||(strcmp(chrLast,locs[0].chromo)!=0)){
    fai = fai_load(fasta);
    chrLast =strdup(locs[0].chromo);//THORFINN leak.. should fix.
    ref = faidx_fetch_seq(fai, chrLast, 0, 0x7fffffff, &ref_len);
    if(ref==NULL){
      fprintf(stderr,"Problems loading chr:%s from fasta\n",locs[0].chromo);
      //      exit(0);
    }
  }

  for(int i=0;i<lens;i++){
    if(strcmp(locs[i].chromo,chrLast)!=0){
      delete [] ref;
      chrLast=locs[i].chromo;
      ref=NULL;
      ref = faidx_fetch_seq(fai, chrLast, 0, 0x7fffffff, &ref_len);
      if(ref==NULL){
	fprintf(stderr,"[%s] Error loading reference alleles\n",__FUNCTION__);
	//exit(0);
      }
    }
    //ret[i] = refToInt[ref[locs[i].position-1]];//offseting by one, to make in work like samtools
    if(ref!=NULL)
      ret[i] = refToInt[ref[locs[i].position]];//offseting by one, to make in work like samtools
    else
      ret[i] = 4;
  }

  return ret;
}


char *getRef2(const loci *locs,int lens,const char *fasta,int doCleanup){
  static char *chrLast =NULL;
  static char *ref = NULL;
  static faidx_t *fai = NULL;
  static int ref_len = -1;
  if(doCleanup){
    if(fai!=NULL)
      fai_destroy(fai);
    free(chrLast);
    free(ref);
    return NULL;
  }


  //  fprintf(stderr,"\t-> Reading fasta ref/anc [%s] fname=%s\n",__FUNCTION__,fasta);

  char *ret=new char[lens];

 
  if(chrLast==NULL||(strcmp(chrLast,locs[0].chromo)!=0)){
    fai = fai_load(fasta);
    chrLast =strdup(locs[0].chromo);//THORFINN leak.. should fix.
    ref = faidx_fetch_seq(fai, chrLast, 0, 0x7fffffff, &ref_len);
  }
  for(int i=0;i<lens;i++){
    if(strcmp(locs[i].chromo,chrLast)!=0){
      delete [] ref;
      chrLast=locs[i].chromo;
      ref=NULL;
      ref = faidx_fetch_seq(fai, chrLast, 0, 0x7fffffff, &ref_len);
      if(ref==NULL){
	fprintf(stderr,"[%s] Error loading reference alleles, will use N/n for entire chromosome\n",__FUNCTION__);
	//	exit(0);
      }
    }
    if(ref!=NULL)
      ret[i] = refToInt[ref[locs[i].position]];//offseting by one, to make in work like samtools
    else
      ret[i] = 4;

  }

  return ret;
}

class getFasta{
private:
  char *ancName;
  char *refName;
  void getRef();
  void getAnc();
public:

  getFasta(argStruct arguments);
  ~getFasta();
  void getOptions(argStruct arguments);
  void run(funkyPars  *pars);//plugs in reference and ancestral
  void printAndClean(funkyPars *pars);

};

getFasta::~getFasta(){
  getRef1(NULL,0,NULL,1);//<-last indicate a cleanup
  getRef2(NULL,0,NULL,1);//<-last indicate a cleanup
}


void getFasta::getOptions(argStruct arguments){
  //default
  refName = NULL;
  ancName = NULL;
  //from command line
  ancName = angsd::getArg("-anc",ancName,arguments);
  refName = angsd::getArg("-ref",refName,arguments);
  //  fprintf(stderr,"refName=%s\tancName=%s\n",refName,ancName);
}

 getFasta::getFasta(argStruct arguments){
  getOptions(arguments);
}

void getFasta::run(funkyPars *pars){
  if(refName!=NULL)
    pars->ref = getRef1(pars->sites,pars->numSites,refName,0);
  if(ancName!=NULL)
    pars->anc = getRef2(pars->sites,pars->numSites,ancName,0);
  
}
